// ForGet. A tool for managing development dependencies.
// Copyright (C) 2015  Daniel Hughes
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
using System;
using System.Xml.Linq;
using System.IO;
using System.Linq;

namespace ForGet
{
    public class Restorer
    {
        readonly string _dependenciesFolder;
        const string _dependenciesFolderName = "dependencies";
        const string _versionFile = "version.txt";
        readonly string _basePath;

        public Restorer(string basePath)
        {
            _basePath = basePath;
            _dependenciesFolder = Path.Combine(basePath, _dependenciesFolderName);
        }

        public bool Restore(string configFile)
        {
            if (!File.Exists(configFile))
            {
                Console.Error.WriteLine("Error: Config file could not be found: {0}", configFile);
                return false;
            }
            if (!Directory.Exists(_dependenciesFolder))
            {
                Directory.CreateDirectory(_dependenciesFolder);
            }
            XElement repo = XElement.Parse(File.ReadAllText(configFile));
            
            var packageSource = ConfigHelpers.GetPackageSource(repo);

            if(packageSource == null)
            {
                return false;
            }
            foreach (var dependency in repo.Elements("Dependency"))
            {
                if (!RestoreDependency(dependency, packageSource))
                {
                    return false;
                }
            }
            return true;
        }

        bool RestoreDependency(XElement dependency, IPackageSource repo)
        {
            string id = dependency.Attribute("Id").Value;
            string version = dependency.Attribute("Version").Value;

            string toPath = Path.Combine(_dependenciesFolder, id);

            if (Directory.Exists(toPath))
            {
                if (File.Exists(Path.Combine(toPath, _versionFile)) &&
                    File.ReadAllText(Path.Combine(toPath, _versionFile)) == version)
                {
                    Console.WriteLine("{0} {1} Already upto date.", id, version);
                    foreach (var copy in dependency.Elements("Copy"))
                    {
                        DoCopyIfMissing(copy, toPath);
                    }
                    return true;
                }
                Console.WriteLine("{0} Removing old version.", id, version);
                Directory.Delete(toPath, true);
            }
            Console.WriteLine("{0} {1} Restoring.", id, version);

            string pathFrom = repo.DownloadPackage(id, version);
            if (pathFrom == null)
            {
                Console.Error.WriteLine("Error: failed to find package {0} {1}", id, version);
                return false;
            }
            new Archive().Install(pathFrom, toPath);
            File.WriteAllText(Path.Combine(toPath, _versionFile), version);

            foreach (var copy in dependency.Elements("Copy"))
            {
                DoCopy(copy, toPath);
            }
            return true;
        }

        string MakePathNative(string path)
        {
            return path
               .Replace("\\", Path.DirectorySeparatorChar.ToString())
               .Replace("/", Path.DirectorySeparatorChar.ToString());
        }

        bool IsFile(string path)
        {
            FileAttributes fileAttributes = File.GetAttributes(path);
            return
               ((fileAttributes & FileAttributes.Directory) != FileAttributes.Directory);
        }

        void DoCopy(XElement copy, string packagePath)
        {
            string fromPath = Path.Combine(packagePath, MakePathNative(copy.Attribute("From").Value));
            string toPath = Path.Combine(_basePath, MakePathNative(copy.Attribute("To").Value));

            if (IsFile(fromPath))
            {
                File.Copy(fromPath, toPath, true);
                return;
            }
            if (Directory.Exists(toPath))
            {
                Directory.Delete(toPath, true);
            }

            //Now Create all of the directories
            Directory.CreateDirectory(toPath);
            foreach (string dirPath in Directory.GetDirectories(fromPath, "*", SearchOption.AllDirectories))
            {
                Directory.CreateDirectory(dirPath.Replace(fromPath, toPath));
            }

            //Copy all the files & Replaces any files with the same name
            foreach (string newPath in Directory.GetFiles(fromPath, "*.*", SearchOption.AllDirectories))
            {
                File.Copy(newPath, newPath.Replace(fromPath, toPath), true);
            }
        }

        void DoCopyIfMissing(XElement copy, string packagePath)
        {
            string fromPath = Path.Combine(packagePath, MakePathNative(copy.Attribute("From").Value));
            string toPath = Path.Combine(_basePath, MakePathNative(copy.Attribute("To").Value));

            if (IsFile(fromPath) && !File.Exists(toPath))
            {
                File.Copy(fromPath, toPath, true);
                Console.WriteLine($"Restoring missing file at {toPath}");
                return;
            }

            //Now Create all of the directories
            if (!Directory.Exists(toPath))
            {
                Console.WriteLine($"Restoring missing folder at {toPath}");
                Directory.CreateDirectory(toPath);
            }
            foreach (string dirPath in Directory.GetDirectories(fromPath, "*", SearchOption.AllDirectories))
            {
                string destPath = dirPath.Replace(fromPath, toPath);
                if (!Directory.Exists(destPath))
                {
                    Console.WriteLine($"Restoring missing folder at {toPath}");
                    Directory.CreateDirectory(destPath);
                }
            }

            //Copy all the files & Replaces any files with the same name
            foreach (string newPath in Directory.GetFiles(fromPath, "*.*", SearchOption.AllDirectories))
            {
                string destPath = newPath.Replace(fromPath, toPath);
                if (!File.Exists(destPath))
                {
                    Console.WriteLine($"Restoring missing file at {toPath}");
                    File.Copy(newPath, destPath, true);
                }
            }
        }
    }
}

